@@ -1,6 +1,6 @@
 /* command.c - Command input and basic commands
  *
- * Copyright (C) 2004-2005 Oskar Liljeblad
+ * Copyright (C) 2004, 2005 Oskar Liljeblad
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -12,9 +12,9 @@
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU Library General Public License for more details.
  *
- * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
+ * You should have received a copy of the GNU General Public License along
+ * with this program; if not, write to the Free Software Foundation,
+ * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
  */
 
 #include <config.h>
@@ -49,6 +49,7 @@
 #include "common/optparser.h"
 #include "microdc.h"
 
+static void cmd_help(int argc, char **argv);
 static void cmd_exit(int argc, char **argv);
 static void cmd_say(int argc, char **argv);
 static void cmd_msg(int argc, char **argv);
@@ -74,10 +75,11 @@ static void cmd_unsearch(int argc, char **argv);
 static void cmd_alias(int argc, char **argv);
 static void cmd_unalias(int argc, char **argv);
 static void cmd_shell(int argc, char **argv);
-static void shell_completion_selector(const char *line,int ws,int we,char *word,PtrV *results);
+static void shell_command_completion_selector(const char *line,int ws,int we,char *word,PtrV *results);
 static void executable_completion_generator(const char *line,int ws,int we,char *word,PtrV *results);
 static void alias_completion_generator(const char *line,int ws,int we,char *word,PtrV *results);
 static void command_completion_generator(const char *line,int ws,int we,char *word,PtrV *results);
+static void alias_command_completion_selector(const char *line,int ws,int we,char *word,PtrV *results);
 
 typedef enum {
     DC_CMD_BUILTIN,
@@ -102,18 +104,22 @@ struct _DCBuiltinCommand {
     DCCommand cmd;
     DCBuiltinCommandHandler handler;
     DCCompletorFunction completor;
+    const char *usage_msg;
+    const char *help_msg;
 };
 
 static TMap *commands; /* initialized in command_init */
 
 static void
-add_builtin_command(const char *name, DCBuiltinCommandHandler handler, DCCompletorFunction completor)
+add_builtin_command(const char *name, DCBuiltinCommandHandler handler, DCCompletorFunction completor, const char *usage_msg, const char *help_msg)
 {
     DCBuiltinCommand *builtin = xmalloc(sizeof(DCBuiltinCommand));
     builtin->cmd.name = xstrdup(name); /* XXX: this should be avoidable (pass char *name instead) */
     builtin->cmd.type = DC_CMD_BUILTIN;
     builtin->handler = handler;
     builtin->completor = completor;
+    builtin->usage_msg = usage_msg;
+    builtin->help_msg = help_msg;
     tmap_put(commands, builtin->cmd.name, builtin);
 }
 
@@ -150,32 +156,131 @@ command_init(void)
 {
     commands = tmap_new();
     tmap_set_compare_fn(commands, (comparison_fn_t) strcmp);
-    add_builtin_command("browse", cmd_browse, user_or_myself_completion_generator);
-    add_builtin_command("cancel", cmd_cancel, transfer_completion_generator);
-    add_builtin_command("cd", cmd_cd, remote_dir_completion_generator);
-    add_builtin_command("connect", cmd_connect, NULL);
-    add_builtin_command("disconnect", cmd_disconnect, NULL);
-    add_builtin_command("exit", cmd_exit, NULL);
-    add_builtin_command("find", cmd_find, remote_path_completion_generator);
-    add_builtin_command("get", cmd_get, remote_path_completion_generator);
-    add_builtin_command("grantslot", cmd_grantslot, user_completion_generator);
-    add_builtin_command("ls", cmd_ls, remote_path_completion_generator);
-    add_builtin_command("msg", cmd_msg, user_completion_generator);
-    add_builtin_command("pwd", cmd_pwd, NULL);
-    add_builtin_command("queue", cmd_queue, user_with_queue_completion_generator);
-    add_builtin_command("raw", cmd_raw, NULL);
-    add_builtin_command("results", cmd_results, NULL);
-    add_builtin_command("say", cmd_say, say_user_completion_generator);
-    add_builtin_command("search", cmd_search, NULL);
-    add_builtin_command("set", cmd_set, set_completion_selector);
-    add_builtin_command("status", cmd_status, NULL);
-    add_builtin_command("transfers", cmd_transfers, NULL);
-    add_builtin_command("unqueue", cmd_unqueue, user_completion_generator);
-    add_builtin_command("unsearch", cmd_unsearch, NULL);
-    add_builtin_command("who", cmd_who, user_completion_generator);
-    add_builtin_command("alias", cmd_alias, NULL);
-    add_builtin_command("unalias", cmd_unalias, alias_completion_generator);
-    add_builtin_command("shell", cmd_shell, shell_completion_selector);
+    /* Note: The algorithm in cmd_help that displays the help message
+     * requires that the help message is terminated by a newline.
+     * Otherwise the last line will not be shown.
+     */
+    add_builtin_command("browse", cmd_browse, user_or_myself_completion_generator,
+        _("browse [USER]"),
+        _("If USER is specified, queue the file list for that user for download and "
+          "start browsing the user's files as soon as the list is downloaded. With no "
+          "arguments, stop browsing.\n"));
+    add_builtin_command("cancel", cmd_cancel, transfer_completion_generator,
+        _("cancel CONNECTION ..."),
+        _("Close a user connection. Use the `transfers' command to get a list of "
+          "connections.\n"));
+    add_builtin_command("cd", cmd_cd, remote_dir_completion_generator,
+        _("cd [DIRECTORY]"),
+        _("Change directory when browsing another user's files. If DIRECTORY is not "
+          "specified, change to the root directory (`/').\n"));
+    add_builtin_command("connect", cmd_connect, NULL,
+        _("connect HOST[:PORT]"),
+        _("Connect to a hub. If PORT is not specified, assume port 411.\n"));
+    add_builtin_command("disconnect", cmd_disconnect, NULL,
+        _("disconnect"),
+        _("Disconnect from the hub.\n"));
+    add_builtin_command("exit", cmd_exit, NULL,
+        _("exit"),
+        _("Quit the program.\n"));
+    add_builtin_command("find", cmd_find, remote_path_completion_generator,
+        _("find [FILE]"),
+        _("List files and directories recursively. Assume current directory if FILE is "
+          "not specified. Must be browsing a user's files to use this command.\n"));
+    add_builtin_command("get", cmd_get, remote_path_completion_generator,
+        _("get FILE ..."),
+        _("Queue file for download. Must be browsing a user's files to use this "
+          "command.\n"));
+    add_builtin_command("grantslot", cmd_grantslot, user_completion_generator,
+        _("grantslot USER..."),
+        _("Grant a download slot for the specified users.\n"));
+    add_builtin_command("help", cmd_help, command_completion_generator,
+        _("help [COMMAND ...]"),
+        _("If COMMAND is specified, display help for that command. Otherwise list all "
+          "available commands.\n"));
+    add_builtin_command("ls", cmd_ls, remote_path_completion_generator,
+        _("ls [OPTION...] [FILE...]"),
+        _("List files and directories. Assume current directory if FILE is not\n"
+          "specified.\n"
+          "\n"
+          "Options:\n"
+          "  -l, --long    use a long listing format\n"));
+    add_builtin_command("msg", cmd_msg, user_completion_generator,
+        _("msg USER MESSAGE..."),
+        _("Send a private message to USER. Note that characters such as semicolon "
+          "(`;'), double quote (`\"') and number sign (`#') in MESSAGE need to be "
+          "escaped or quoted. Therefore it is recommended to put MESSAGE in double "
+          "quotes.\n"
+          "\n"
+          "Example:\n"
+          "  msg some_user \"hello, how are you?\"\n"));
+    add_builtin_command("pwd", cmd_pwd, NULL,
+        _("pwd"),
+        _("Display user being browsed and current directory.\n"));
+    add_builtin_command("queue", cmd_queue, user_with_queue_completion_generator,
+        _("queue USER ..."),
+        _("Display files queued for download from the specified users.\n"));
+    add_builtin_command("raw", cmd_raw, NULL,
+        _("raw DATA..."),
+        _("Send some raw data to the hub. Note that characters such as semicolon "
+          "(`;'), double quote (`\"') and number sign (`#') in DATA need to be "
+          "escaped or quoted. Therefore it is recommended to put DATA in double "
+          "quotes.\n"));
+    add_builtin_command("results", cmd_results, NULL,
+        _("results [INDEX ...]"),
+        _("If INDEX is specified, display results for the search by that index. "
+          "Otherwise, display a list of searches and statistics over those searches.\n"));
+    add_builtin_command("say", cmd_say, say_user_completion_generator,
+        _("say MESSAGE..."),
+        _("Send a public message to users on the hub. Note that characters such as "
+          "semicolon (`;'), double quote (`\"') and number sign (`#') in MESSAGE need "
+          "to be escaped or quoted. Therefore it is recommended to put MESSAGE in "
+          "double quotes.\n"
+          "\n"
+          "Example:\n"
+          "  say \"hi everyone!\"\n"));
+    add_builtin_command("search", cmd_search, NULL,
+        _("search WORD..."),
+        _("Issue a search for the specified search words.\n"));
+    add_builtin_command("set", cmd_set, set_command_completion_selector,
+        _("set [NAME [VALUE...]]"),
+        _("Without arguments, display a list of variables and their current values. With "
+          "only NAME argument, display the value of that variable. With NAME and VALUE "
+          "arguments, change the value of a variable.\n"));
+    add_builtin_command("status", cmd_status, NULL,
+        _("status"),
+        _("Display status information and some statistics.\n"));
+    add_builtin_command("transfers", cmd_transfers, NULL,
+        _("transfers"),
+        _("Display a list of user connections.\n"));
+    add_builtin_command("unqueue", cmd_unqueue, user_completion_generator,
+        _("unqueue USER [RANGE]"),
+        _("Remove all or a range of queued files for USER. If RANGE is not specified, "
+          "remove all files from the queue. Use dash (`-') and comma (`,') in RANGE. "
+          "Open ranges are accepted (e.g. `1-' or `-2').\n"));
+    add_builtin_command("unsearch", cmd_unsearch, NULL,
+        _("unsearch INDEX ..."),
+        _("Remove a previously issued search and all results for that search.\n"));
+    add_builtin_command("who", cmd_who, user_completion_generator,
+        _("who [USER ...]"),
+        _("If USER is specified, display information on that user. Otherwise, "
+          "display a table of users with some user details.\n"));
+    add_builtin_command("alias", cmd_alias, alias_command_completion_selector,
+        _("alias [NAME[=VALUE] ...]"),
+        _("Without arguments, display the list of aliases. With NAME argument, "
+          "display what value (command) that alias is set to. With both NAME and "
+          "VALUE argument, change alias. Note that VALUE is a single argument - "
+          "you need to use quotes for more complex commands.\n"
+          "\n"
+          "Example:\n"
+          "  alias ll \"ls -l\"\n"));
+    add_builtin_command("unalias", cmd_unalias, alias_completion_generator,
+        _("unalias NAME ..."),
+        _("Remove aliases.\n"));
+    add_builtin_command("shell", cmd_shell, shell_command_completion_selector,
+        _("shell COMMAND [ARGUMENTS...]"),
+        _("Execute a system command. microdc will continue in the background while "
+          "the command is executing.\n"));
+
     add_alias_command("ll", "ls -l");
 }
 
@@ -222,6 +327,84 @@ default_completion_selector(const char *line,int ws,int we,char *word,PtrV *resu
     }
 }
 
+static char *
+replace_substr(const char *str, int sp, int ep, const char *repl)
+{
+    char *out;
+    size_t len;
+    size_t repl_len;
+
+    len = strlen(str);
+    repl_len = strlen(repl);
+    out = xmalloc(sp + strlen(repl) + len - ep + 1);
+    memcpy(out, str, sp);
+    memcpy(out+sp, repl, repl_len);
+    memcpy(out+sp+repl_len, str+ep, len-ep+1);
+
+    return out;
+}
+
+
+static void
+alias_command_completion_selector(const char *line,int ws,int we,char *word,PtrV *results)
+{
+    char *newline;
+    char *newword;
+    int newwe;
+    int newws;
+    bool quoted;
+
+    /*printf("\nOLD: %2d:%2d [%s:%s]\n", ws, we, word, line);*/
+    newline = word;
+    newwe = strlen(word);
+    newws = find_completion_word_start(newline, newwe);
+    /*newword = dequote_word(newline+newws, false, NULL);*/
+    newword = get_completion_word_dequoted(newline, newwe);
+//    printf("NEW: %2d:%2d [%s:%s]\n", newws, newwe, newword, newline);
+
+    quoted = newline[newws] == '"';
+    default_completion_selector(newline, newws, newwe, newword, results);
+    if (results->cur == 1) {
+        DCCompletionEntry *entry = results->buf[0];
+        char *match;
+        char *match2;
+
+        if (entry->input_single_full_format != NULL && strcmp(newword, entry->input) == 0) {
+            match = xasprintf(entry->input_single_full_format, entry->input);
+        } else if (entry->input_single_format != NULL) {
+            match = xasprintf(entry->input_single_format, entry->input);
+        } else {
+            match = xasprintf(entry->input_format, entry->input);
+        }
+        if (entry->input != entry->display)
+            free(entry->input);
+        match2 = quote_word_full(match, quoted, true, ";", "#", false, true, true, true);
+        free(match);
+        entry->input = replace_substr(newline, newws, newwe, match2);
+//        printf(">>[%s]\n", entry->input);
+        free(match2);
+        entry->input_format = "%s";
+        entry->input_single_format = NULL;
+        entry->input_single_full_format = NULL;
+        entry->inhibit_quote = true;
+    } else if (results->cur > 1) {
+        uint32_t c;
+
+        for (c = 0; c < results->cur; c++) {
+            DCCompletionEntry *entry = results->buf[c];
+            char *match;
+
+            match = quote_word_full(entry->input, quoted, true, ";", "#", false, true, true, true);
+            if (entry->input != entry->display)
+                free(entry->input);
+            entry->input = replace_substr(newline, newws, newwe, match);
+            free(match);
+        }
+    }
+
+    free(newword);
+}
+
 static void
 alias_completion_generator(const char *line,int ws,int we,char *word,PtrV *results)
 {
@@ -376,7 +559,7 @@ executable_completion_generator(const char *line,int ws,int we,char *word,PtrV *
 }
 
 static void
-shell_completion_selector(const char *line,int ws,int we,char *word,PtrV *results)
+shell_command_completion_selector(const char *line,int ws,int we,char *word,PtrV *results)
 {
     if (get_word_index(line, ws) == 1) {
         executable_completion_generator(line,ws,we,word,results);
@@ -385,6 +568,79 @@ shell_completion_selector(const char *line,int ws,int we,char *word,PtrV *result
     }
 }
 
+static void
+cmd_help(int argc, char **argv)
+{
+    uint32_t c;
+    int width;
+
+    screen_get_size(NULL, &width);
+
+    if (argc == 1) {
+        TMapIterator it;
+        StrBuf *buf;
+
+        buf = strbuf_new();
+        for (tmap_iterator(commands, &it); it.has_next(&it); ) {
+            DCCommand *cmd = it.next(&it);
+            if (cmd->type == DC_CMD_BUILTIN) {
+                DCBuiltinCommand *builtin = (DCBuiltinCommand *) cmd;
+                if (strlen(builtin->usage_msg) < width/2) {
+                    if (strbuf_is_empty(buf)) {
+                        strbuf_append(buf, builtin->usage_msg);
+                        strbuf_append_char_n(buf, width/2-strbuf_length(buf), ' ');
+                    } else {
+                        screen_putf("%s%s\n", strbuf_buffer(buf), builtin->usage_msg);
+                        strbuf_clear(buf);
+                    }
+                } else {
+                    if (strbuf_is_empty(buf)) {
+                        screen_putf("%s\n", builtin->usage_msg);
+                    } else {
+                        screen_putf("%s\n%s\n", strbuf_buffer(buf), builtin->usage_msg);
+                        strbuf_clear(buf);
+                    }
+                }
+            }
+        }
+        if (!strbuf_is_empty(buf))
+            screen_putf("%s\n", strbuf_buffer(buf));
+        strbuf_free(buf);
+        return;
+    }
+
+
+    for (c = 1; c < argc; c++) {
+        DCCommand *cmd;
+        cmd = tmap_get(commands, argv[c]);
+        if (cmd == NULL) {
+            warn("%s: Unknown command.\n", quotearg(argv[c]));
+        } else {
+            if (cmd->type == DC_CMD_ALIAS) {
+                DCAliasCommand *alias = (DCAliasCommand *) cmd;
+                screen_putf("%s: aliased to `%s'.\n", quotearg_n(0, cmd->name), quotearg_n(1, alias->alias_spec));
+            } else {
+                DCBuiltinCommand *builtin = (DCBuiltinCommand *) cmd;
+                const char *p0;
+                const char *p1;
+
+                screen_putf("%s: %s\n", cmd->name, builtin->usage_msg);
+                for (p0 = builtin->help_msg; (p1 = strchr(p0, '\n')) != NULL; p0 = p1+1) {
+                    PtrV *msgs;
+                    uint32_t d;
+
+                    msgs = wordwrap(p0, p1-p0, width-5-4, width-5-4);
+                    for (d = 0; d < msgs->cur; d++)
+                        screen_putf("    %s\n", (char *) msgs->buf[d]);
+                    if (d == 0)
+                        screen_putf("\n");
+                    ptrv_free(msgs);
+                }
+            }
+        }
+    }
+}
+
 static void
 cmd_shell(int argc, char **argv)
 {
@@ -395,8 +651,9 @@ cmd_shell(int argc, char **argv)
 	return;
     }
     if (shell_child == 0) {
+        warn_writer = default_warn_writer;
         if (execvp(argv[1], argv+1) < 0)
-	    die(_("Cannot execute - %s\n"), errstr); /* die OK */
+	    die(_("%s: cannot execute - %s\n"), quotearg(argv[1]), errstr); /* die OK */
         exit(EXIT_FAILURE); /* shouldn't get here */
     }
 }
@@ -816,7 +1073,7 @@ cmd_ls(int argc, char **argv)
 {
     OptDetail long_opts[] = {
 	{ "l", OPTP_NO_ARG, 'l' },
-	{ "help", OPTP_NO_ARG, 'h' },
+/*	{ "help", OPTP_NO_ARG, 'h' },*/
 	{ NULL },
     };
     OptParser *p;
@@ -829,12 +1086,12 @@ cmd_ls(int argc, char **argv)
 	case 'l':
 	    long_mode = true;
 	    break;
-	case 'h':
+	/*case 'h':
 	    screen_putf("Usage: ls [OPTION]... [FILE]...\n");
 	    screen_putf("  -l          use a long listing format\n");
 	    screen_putf("      --help  display this help and exit\n");
 	    optparser_free(p);
-	    return;
+	    return;*/
 	}
     }
     if (optparser_error(p) != NULL) {
@@ -894,11 +1151,7 @@ append_download_dir(DCUserInfo *ui, DCFileList *node, char *base_path, uint64_t
 static void
 cmd_get(int argc, char **argv)
 {
-    DCFileList *node;
-    char *path;
-    const char *file_name;
-    uint64_t bytes = 0;
-    uint32_t files = 0;
+    uint32_t c;
 
     /* Note: This command assumes that we are browsing some user.
      * If we in the future would allow browse-less getting,
@@ -906,9 +1159,8 @@ cmd_get(int argc, char **argv)
      * found in cmd_browse, such as strcmp(my_nick, nick)==0 etc.
      */
 
-    file_name = argv[1];
-    if (file_name == NULL) {
-    	screen_putf(_("Usage: %s FILE\n"), argv[0]);
+    if (argc == 1) {
+    	screen_putf(_("Usage: %s FILE ...\n"), argv[0]);
 	return;
     }
     if (browse_list == NULL) {
@@ -920,35 +1172,40 @@ cmd_get(int argc, char **argv)
 	return;
     }
 
-    path = apply_cwd(file_name);
-    node = filelist_lookup(browse_list, path);
-    if (node == NULL) {
-	screen_putf(_("%s: No such file or directory\n"), quotearg(argv[1]));
-	free(path);
-	return;
-    }
-    free(path);
+    for (c = 1; c < argc; c++) {
+        DCFileList *node;
+        char *path;
+        uint64_t bytes = 0;
+        uint32_t files = 0;
 
-    if (append_download_dir(browse_user, node, browse_path, &bytes, &files)) {
-        char *fmt1 = ngettext("byte", "bytes", bytes);
-        char *fmt2 = ngettext("file", "files", files);
-    	path = filelist_get_path(node);
-	screen_putf(_("Downloading %s (%" PRId64 " %s in %d %s)\n"), quotearg(path), bytes, fmt1, files, fmt2);
-	free(path);
-    } else {
-        /* XXX: hmm, shouldn't we return even if queue is non-empty? */
-    	if (browse_user->download_queue->cur == 0) {
-	    screen_putf(_("No files to download (empty directories).\n"));
-	    return;
-	}
+        path = apply_cwd(argv[c]);
+        node = filelist_lookup(browse_list, path);
+        if (node == NULL) {
+            screen_putf(_("%s: No such file or directory\n"), quotearg(argv[c]));
+            free(path);
+            return;
+        }
+        free(path);
+
+        if (append_download_dir(browse_user, node, browse_path, &bytes, &files)) {
+            char *fmt1 = ngettext("byte", "bytes", bytes);
+            char *fmt2 = ngettext("file", "files", files);
+            path = filelist_get_path(node);
+            screen_putf(_("Downloading %s (%" PRId64 " %s in %d %s)\n"), quotearg(path), bytes, fmt1, files, fmt2); /* " */
+            free(path);
+        } else {
+            /* XXX: hmm, shouldn't we return even if queue is non-empty? */
+            if (browse_user->download_queue->cur == 0) {
+                screen_putf(_("%s: No files to download (empty directories).\n"), quotearg(argv[c]));
+                continue;
+            }
+        }
     }
 
     if (!has_user_conn(browse_user, DC_DIR_RECEIVE) && browse_user->conn_count < DC_USER_MAX_CONN) {
-    	hub_connect_user(browse_user); /* Ignore errors */
+        hub_connect_user(browse_user); /* Ignore errors */
     } else {
-	screen_putf(
-          ngettext("No free connections. Queued file for download.\n",
-	  	   "No free connections. Queued files for download.\n", files));
+        screen_putf(_("No free connections. Queued files for download.\n"));
     }
 }
 
@@ -956,8 +1213,7 @@ cmd_get(int argc, char **argv)
 static void
 cmd_queue(int argc, char **argv)
 {
-    uint32_t c;
-    DCUserInfo *ui;
+    uint32_t d;
 
     if (argc == 1) {
     	screen_putf(_("Usage: %s USER\n"), argv[0]);
@@ -967,17 +1223,23 @@ cmd_queue(int argc, char **argv)
     	screen_putf(_("Not connected.\n"));
 	return;
     }
-    ui = hmap_get(hub_users, argv[1]);
-    if (ui == NULL) {
-        screen_putf(_("%s: No such user on this hub\n"), quotearg(argv[1]));
-	return;
-    }
-    for (c = 0; c < ui->download_queue->cur; c++) {
-	DCQueuedFile *queued = ui->download_queue->buf[c];
-	screen_putf(_("%d. [ %s ] %s\n"),
-		c+1,
-		quotearg_n(0, queued->base_path),
-		quotearg_n(1, queued->filename + strlen(queued->base_path)));
+    for (d = 1; d < argc; d++) {
+        uint32_t c;
+        DCUserInfo *ui;
+
+        ui = hmap_get(hub_users, argv[d]);
+        if (ui == NULL) {
+            screen_putf(_("%s: No such user on this hub\n"), quotearg(argv[d]));
+            continue;
+        }
+        screen_putf("%s:\n", quotearg(ui->nick));
+        for (c = 0; c < ui->download_queue->cur; c++) {
+            DCQueuedFile *queued = ui->download_queue->buf[c];
+            screen_putf(_("%d. [ %s ] %s\n"),
+                    c+1,
+                    quotearg_n(0, queued->base_path),
+                    quotearg_n(1, queued->filename + strlen(queued->base_path)));
+        }
     }
 }
 
@@ -1078,7 +1340,7 @@ cmd_who(int argc, char **argv)
     uint32_t c;
     DCUserInfo **items;
     uint32_t count;
-    uint32_t cols;
+    int cols;
     StrBuf *out;
 
     if (hub_state < DC_HUB_LOGGED_IN) {
@@ -1087,22 +1349,24 @@ cmd_who(int argc, char **argv)
     }
 
     if (argc > 1) {
-    	DCUserInfo *ui;
+        for (c = 1; c < argc; c++) {
+            DCUserInfo *ui;
 
-	ui = hmap_get(hub_users, argv[1]);
-	if (ui == NULL) {
-    	    screen_putf(_("%s: No such user on this hub\n"), quotearg(argv[1]));
-	} else {
-	    char *fmt1 = ngettext("byte", "bytes", ui->share_size);
-    	    screen_putf(_("Nick: %s\n"), quotearg(ui->nick));
-    	    screen_putf(_("Description: %s\n"), quotearg(IFNULL(ui->description, "")));
-    	    screen_putf(_("Speed: %s\n"), quotearg(IFNULL(ui->speed, "")));
-    	    screen_putf(_("Level: %d\n"), ui->level);
-    	    screen_putf(_("E-mail: %s\n"), quotearg(IFNULL(ui->email, "")));
-    	    screen_putf(_("Operator: %d\n"), ui->is_operator);
-    	    screen_putf(_("Share Size: %" PRId64 " %s (%" PRId64 " MB)\n"),
-    	        ui->share_size, fmt1, ui->share_size/(1024*1024));
-	}
+            ui = hmap_get(hub_users, argv[c]);
+            if (ui == NULL) {
+                screen_putf(_("%s: No such user on this hub\n"), quotearg(argv[c]));
+            } else {
+                char *fmt1 = ngettext("byte", "bytes", ui->share_size);
+                screen_putf(_("Nick: %s\n"), quotearg(ui->nick));
+                screen_putf(_("Description: %s\n"), quotearg(IFNULL(ui->description, "")));
+                screen_putf(_("Speed: %s\n"), quotearg(IFNULL(ui->speed, "")));
+                screen_putf(_("Level: %d\n"), ui->level);
+                screen_putf(_("E-mail: %s\n"), quotearg(IFNULL(ui->email, "")));
+                screen_putf(_("Operator: %d\n"), ui->is_operator);
+                screen_putf(_("Share Size: %" PRId64 " %s (%" PRId64 " MB)\n"), /* " */
+                    ui->share_size, fmt1, ui->share_size/(1024*1024));
+            }
+        }
     	return;
     }
 
@@ -1129,7 +1393,7 @@ cmd_who(int argc, char **argv)
 	strbuf_clear(out);
 	strbuf_append(out, nick);
 	strbuf_append_char_n(out, maxlen+1-strlen(nick), ' ');
-	strbuf_appendf(out, "  %7" PRId64 "M", ui->share_size / (1024*1024));
+	strbuf_appendf(out, "  %7" PRId64 "M", ui->share_size / (1024*1024)); /* " */
 	strbuf_append(out, ui->is_operator ? " op" : "   ");
 	if (ui->download_queue->cur > 0)
 	    strbuf_appendf(out, " (%3d)", ui->download_queue->cur);
@@ -1185,20 +1449,24 @@ cmd_transfers(int argc, char **argv)
 static void
 cmd_cancel(int argc, char **argv)
 {
-    DCUserConn *uc;
+    uint32_t c;
 
     if (argc == 1) {
-    	screen_putf(_("Usage: %s CONNECTION\n"), argv[0]);
+    	screen_putf(_("Usage: %s CONNECTION ...\n"), argv[0]);
 	return;
     }
 
-    uc = hmap_get(user_conns, argv[1]);
-    if (uc == NULL) {
-    	screen_putf(_("%s: No such user connection.\n"), quotearg(argv[1]));
-	return;
+    for (c = 1; c < argc; c++) {
+        DCUserConn *uc;
+
+        uc = hmap_get(user_conns, argv[c]);
+        if (uc == NULL) {
+    	    screen_putf(_("%s: No such user connection.\n"), quotearg(argv[c]));
+        } else {
+            user_conn_cancel(uc);
+        } 
     }
 
-    user_conn_cancel(uc);
 }
 
 static void
@@ -1223,7 +1491,7 @@ cmd_search(int argc, char **argv)
 static void
 cmd_results(int argc, char **argv)
 {
-    uint32_t c;
+    uint32_t d;
 
     if (argc == 1) {
         time_t now;
@@ -1233,63 +1501,75 @@ cmd_results(int argc, char **argv)
             return;
         }
 
-    	for (c = 0; c < our_searches->cur; c++) {
-    	    DCSearchRequest *sd = our_searches->buf[c];
+    	for (d = 0; d < our_searches->cur; d++) {
+    	    DCSearchRequest *sd = our_searches->buf[d];
     	    char *status;
     	    char *spec;
 
             spec = search_selection_to_string(&sd->selection);
             status = sd->issue_time + SEARCH_TIME_THRESHOLD <= now
                    ? _("Closed") : _("Open");
-	    screen_putf(_("%d. %s (%s) Results: %d\n"), c+1, quotearg(spec),
+	    screen_putf(_("%d. %s (%s) Results: %d\n"), d+1, quotearg(spec),
                    status, sd->responses->cur);
 	}
-    } else {
-    	DCSearchRequest *sd;
-	if (!parse_uint32(argv[1], &c) || c == 0 || c-1 >= our_searches->cur) {
-    	    screen_putf(_("Invalid search index.\n"));
-	    return;
-	}
-	sd = our_searches->buf[c-1];
-	for (c = 0; c < sd->responses->cur; c++) {
-	    DCSearchResponse *sr = sd->responses->buf[c];
-	    char *n;
-	    char *t;
+	return;
+    }
 
-	    n = translate_remote_to_local(sr->filename);
-	    if (sr->filetype == DC_TYPE_DIR) /* XXX: put into some function */
-	    	t = "/";
-	    else
-	    	t = "";
-	    screen_putf("%d. %s %s%s\n", c+1, quotearg(sr->userinfo->nick), n, t);
-	    free(n);
-	}
+    for (d = 1; d < argc; d++) {
+        DCSearchRequest *sd;
+        uint32_t c;
+
+        if (!parse_uint32(argv[d], &c) || c == 0 || c-1 >= our_searches->cur) {
+            screen_putf(_("%s: Invalid search index.\n"), quotearg(argv[d]));
+            continue;
+        }
+        sd = our_searches->buf[c-1];
+        screen_putf(_("Search %d:\n"), c);
+        for (c = 0; c < sd->responses->cur; c++) {
+            DCSearchResponse *sr = sd->responses->buf[c];
+            char *n;
+            char *t;
+
+            n = translate_remote_to_local(sr->filename);
+            if (sr->filetype == DC_TYPE_DIR) /* XXX: put into some function */
+                t = "/";
+            else
+                t = "";
+            screen_putf("%d. %s %s%s\n", c+1, quotearg(sr->userinfo->nick), n, t);
+            free(n);
+        }
     }
 }
 
 static void
 cmd_unsearch(int argc, char **argv)
 {
-    DCSearchRequest *sd;
-    uint32_t index;
+    uint32_t c;
 
     if (argc == 1) {
     	screen_putf(_("Usage: %s INDEX\n"), argv[0]);
 	return;
     }
-    if (!parse_uint32(argv[1], &index) || index == 0 || index-1 >= our_searches->cur) {
-    	screen_putf(_("Invalid search index %s.\n"), quote(argv[1]));
-	return;
-    }
+    for (c = 1; c < argc; c++) {
+        DCSearchRequest *sd;
+        uint32_t index;
 
-    sd = our_searches->buf[index-1];
-    ptrv_remove_range(our_searches, index-1, index);
-    free_search_request(sd);    
+        if (!parse_uint32(argv[c], &index) || index == 0 || index-1 >= our_searches->cur) {
+            screen_putf(_("%s: Invalid search index.\n"), quotearg(argv[c]));
+            return;
+        }
+
+        sd = our_searches->buf[index-1];
+        ptrv_remove_range(our_searches, index-1, index);
+        free_search_request(sd);    
+    }
 }
 
 static void
 cmd_alias(int argc, char **argv)
 {
+    uint32_t c;
+
     if (argc == 1) {
 	TMapIterator it;
 
@@ -1300,43 +1580,62 @@ cmd_alias(int argc, char **argv)
 		screen_putf("alias %s \"%s\"\n" /*no translation */, cmd->name, quotearg(alias->alias_spec));
 	    }
 	}
-    } else if (argc == 2) {
-	DCCommand *cmd = tmap_get(commands, argv[1]);
-	if (cmd != NULL && cmd->type == DC_CMD_ALIAS) {
-	    DCAliasCommand *alias = (DCAliasCommand *) cmd;
-	    screen_putf("alias %s \"%s\"\n" /*no translation */, argv[1], quotearg(alias->alias_spec));
-	} else {
-	    warn(_("%s: No such alias.\n"), quotearg(argv[1]));
-	}
-    } else if (argc == 3) {
-	DCCommand *cmd = tmap_get(commands, argv[1]);
-	if (cmd == NULL) {
-	    DCAliasCommand *alias = xmalloc(sizeof(DCAliasCommand));
-	    alias->cmd.name = xstrdup(argv[1]);
-	    alias->cmd.type = DC_CMD_ALIAS;
-	    alias->alias_spec = xstrdup(argv[2]);
-	    tmap_put(commands, alias->cmd.name, alias);
-	} else if (cmd->type == DC_CMD_ALIAS) {
-	    DCAliasCommand *alias = (DCAliasCommand *) cmd;
-	    free(alias->alias_spec);
-	    alias->alias_spec = xstrdup(argv[2]);
-	} else {
-	    warn(_("%s: Cannot override built-in command.\n"), quotearg(cmd->name));
-	}
-    } else {
-        warn(_("%s: Too many arguments.\n"), argv[0]);
+	return;
+    }
+
+    for (c = 1; c < argc; c++) {
+        char *name = argv[c];
+        char *value;
+        DCCommand *cmd;
+
+        value = strchr(name, '=');
+        if (value == NULL) {
+            cmd = tmap_get(commands, name);
+            if (cmd != NULL && cmd->type == DC_CMD_ALIAS) {
+                DCAliasCommand *alias = (DCAliasCommand *) cmd;
+                screen_putf("alias %s=\"%s\"\n" /*no translation */, name, quotearg(alias->alias_spec));
+            } else {
+                warn(_("%s: No such alias.\n"), quotearg(name));
+            }
+        } else {
+            *value = '\0';
+            value++;
+            if (strpbrk(name, " \"#;") != NULL) {
+                warn(_("%s: Invalid alias name\n"), quotearg(name));
+                continue;
+            }
+            cmd = tmap_get(commands, name);
+            if (cmd == NULL) {
+                DCAliasCommand *alias = xmalloc(sizeof(DCAliasCommand));
+                alias->cmd.name = xstrdup(name);
+                alias->cmd.type = DC_CMD_ALIAS;
+                alias->alias_spec = xstrdup(value);
+                tmap_put(commands, alias->cmd.name, alias);
+            } else if (cmd->type == DC_CMD_ALIAS) {
+                DCAliasCommand *alias = (DCAliasCommand *) cmd;
+                free(alias->alias_spec);
+                alias->alias_spec = xstrdup(value);
+            } else {
+                warn(_("%s: Cannot override built-in command.\n"), quotearg(cmd->name));
+            }
+        }
     }
 }
 
 static void
 cmd_unalias(int argc, char **argv)
 {
+    uint32_t c;
+
     if (argc == 1) {
-    	screen_putf(_("Usage: %s ALIAS\n"), argv[0]);
-    } else {
-	DCCommand *cmd = tmap_get(commands, argv[1]);
+    	screen_putf(_("Usage: %s NAME ...\n"), argv[0]);
+    	return;
+    }
+
+    for (c = 0; c < argc; c++) {
+	DCCommand *cmd = tmap_get(commands, argv[c]);
 	if (cmd == NULL || cmd->type != DC_CMD_ALIAS) {
-	    warn(_("%s: No such alias.\n"), quotearg(argv[1]));
+	    warn(_("%s: No such alias.\n"), quotearg(argv[c]));
 	} else {
 	    DCAliasCommand *alias = (DCAliasCommand *) cmd;
 	    tmap_remove(commands, cmd->name);
