@@ -12,9 +12,9 @@
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
  *
- * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
+ * You should have received a copy of the GNU General Public License along
+ * with this program; if not, write to the Free Software Foundation,
+ * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
  */
 
 /* TODO: unified handling of whitespace (single point of definition),
@@ -360,6 +360,7 @@ find_unquoted_leading_char(const char *str, const char *maxend, char ch)
 char *
 get_completion_word_dequoted(const char *str, int pos)
 {
+    /*return dequote_words(find_completion_word_start(str, pos), false, str+pos);*/
     const char *maxend = str + pos;
     const char *start = str;
     int index;
@@ -376,6 +377,26 @@ get_completion_word_dequoted(const char *str, int pos)
     }
 }
 
+int
+find_completion_word_start(const char *str, int pos)
+{
+    const char *maxend = str + pos;
+    const char *start = str;
+    const char *end = str;
+    int index;
+
+    for (index = 0; ; index++) {
+        end = find_word_end(end, maxend);
+        if (end >= maxend || *end == '\0') {
+	    start = find_word_start(start, maxend);
+	    if (start >= maxend || *start == '\0')
+		return pos;
+	    return start-str;
+	}
+	start = end;
+    }
+}
+
 /* Split input into an array of words. The array will be null-terminated.
  *
  * XXX: this function has not been tested
@@ -491,3 +512,14 @@ char_is_quoted(char *string, int index)
 
     return 0;
 }
+
+int
+count_unquoted_chars(const char *str, const char *end)
+{
+    int count = 0;
+    for (; *str && str < end; str++) {
+        if (*str != '"' && *str != '\\')
+            count++;
+    }
+    return count;
+}
