@@ -1,6 +1,6 @@
 /* fs.c - Local and remote file system management
  *
- * Copyright (C) 2004-2005 Oskar Liljeblad
+ * Copyright (C) 2004, 2005 Oskar Liljeblad
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -12,9 +12,9 @@
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU Library General Public License for more details.
  *
- * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
+ * You should have received a copy of the GNU General Public License along
+ * with this program; if not, write to the Free Software Foundation,
+ * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
  */
 
 #include <config.h>
@@ -390,7 +390,7 @@ DCFileList *
 filelist_open(const char *filename)
 {
     struct stat st;
-    char *contents;
+    uint8_t *contents;
     char *decoded;
     uint32_t decoded_len;
     int fd;
@@ -532,7 +532,7 @@ filelist_create(const char *basedir)
 
     len = strbuf_length(sb);
     indata = strbuf_free_to_string(sb);
-    outdata = huffman_encode(indata, len, &len);
+    outdata = huffman_encode((uint8_t *) indata, len, &len);
     free(indata);
 
     filename = catfiles(listing_dir, "MyList.DcLst");
@@ -718,6 +718,7 @@ local_fs_completion_generator(const char *line,int ws,int we,char *word,PtrV *re
 		        input = catfiles_with_trailing_slash(dir_part, de->d_name);
 		        display = xasprintf("%s/", de->d_name);
 		        entry->input_char = '\0';
+		        entry->inhibit_quote = true;
                         entry->display = display;
                         entry->input = input;
                         ptrv_append(results, entry);
@@ -830,6 +831,7 @@ remote_fs_completion_generator(const char *line,int ws,int we,char *word,PtrV *r
 		entry->display = xasprintf("%s/", node->name);
 		entry->input = catfiles_with_trailing_slash(dir_part, node->name);
 		entry->input_char = '\0';
+		entry->inhibit_quote = true;
 	    } else {
 	        entry = new_completion_entry(NULL, NULL);
 	        entry->display = xstrdup(node->name);
