@@ -1,6 +1,6 @@
 /* hub.c - Hub communication
  *
- * Copyright (C) 2004-2005 Oskar Liljeblad
+ * Copyright (C) 2004, 2005 Oskar Liljeblad
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -12,9 +12,9 @@
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU Library General Public License for more details.
  *
- * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
+ * You should have received a copy of the GNU General Public License along
+ * with this program; if not, write to the Free Software Foundation,
+ * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
  */
 
 #include <config.h>
@@ -179,7 +179,7 @@ bool
 hub_putf(const char *format, ...)
 {
     va_list args;
-    uint32_t oldcur;
+    size_t oldcur;
     int res;
 
     oldcur = hub_sendq->cur;
@@ -718,7 +718,7 @@ hub_now_writable(void)
 {
     if (hub_state == DC_HUB_CONNECT) {
 	int error;
-        int size = sizeof(error);
+        socklen_t size = sizeof(error);
 	socklen_t addr_len;
 
         if (getsockopt(hub_socket, SOL_SOCKET, SO_ERROR, &error, &size) < 0) {
