@@ -1,6 +1,6 @@
 /* main.c - Main routine and some common functions
  *
- * Copyright (C) 2004-2005 Oskar Liljeblad
+ * Copyright (C) 2004, 2005 Oskar Liljeblad
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -12,9 +12,9 @@
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU Library General Public License for more details.
  *
- * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
+ * You should have received a copy of the GNU General Public License along
+ * with this program; if not, write to the Free Software Foundation,
+ * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
  */
 
 #include <config.h>
@@ -38,6 +38,7 @@
 #include "version-etc.h"	/* Gnulib */
 #include "xalloc.h"		/* Gnulib */
 #include "quotearg.h"		/* Gnulib */
+#include "sig2str.h"		/* Gnulib */
 #include "gettext.h"		/* Gnulib/GNU gettext */
 #define _(s) gettext(s)
 #define N_(s) gettext_noop(s)
@@ -78,7 +79,7 @@ struct in_addr force_listen_addr = { INADDR_NONE };
 /* settings/variables */
 int used_ul_slots = 0;
 int used_dl_slots = 0;
-int my_ul_slots;
+uint32_t my_ul_slots;
 char *my_nick;
 char *my_tag;
 char *my_description;
@@ -110,17 +111,21 @@ PtrV *delete_dirs = NULL;
 static PtrV *search_udpmsg_out;	/* pending outgoing search results */
 pid_t shell_child = -1;
 
-const char version_etc_copyright[] = "Copyright (C) 2004-2005 Oskar Liljeblad";
+/* XXX: change to ("Copyright %s %d Oskar Liljeblad", gettext("(C)"), 2005) */
+const char version_etc_copyright[] = "Copyright (C) 2004, 2005 Oskar Liljeblad";
 
 static uint32_t user_conn_unknown_last = 0;
 static PtrV *user_conn_unknown_free = NULL;
 static ByteQ *search_recvq;
 
+static pid_t main_process_id;
 static int signal_pipe[2] = { -1, -1 };
 static int listen_socket = -1;
 static int search_socket = -1;
 
+static const char *short_opts = "n";
 static struct option long_opts[] = {
+    { "no-config", no_argument, NULL, 'n' },
     { "version", no_argument, NULL, VERSION_OPT },
     { "help", no_argument, NULL, HELP_OPT },
     { 0, 0, 0, 0 }
@@ -491,6 +496,13 @@ signal_received(int signal)
 {
     uint8_t signal_char;
 
+    /* This check is to stop a small risk: The signal may be received in a
+     * newly created child process that hasn't had its signal handler
+     * changed yet.
+     */
+    if (getpid() != main_process_id)
+        return;
+
     signal_char = signal;
     if (write(signal_pipe[1], &signal, sizeof(uint8_t)) < sizeof(uint8_t)) {
 	/* Die only if the signal is fatal.
@@ -530,14 +542,26 @@ read_signal_input(void)
         int status;
 
         while ((child = waitpid(-1, &status, WNOHANG)) > 0) {
+            char *name;
+
             if (child == shell_child) {
                 screen_wakeup(WIFSIGNALED(status) && WTERMSIG(status) == SIGINT);
                 shell_child = -1;
+                name = _("Shell process");
             } else {
                 /* Assume it was a child process */
-                if (!WIFEXITED(status))
-                    warn(_("User process exited with status %d\n"), status);
+                name = _("User process");
             }
+            if (WIFEXITED(status) && WEXITSTATUS(status) != 0) {
+                warn(_("%s exited with return code %d.\n"), name, WEXITSTATUS(status));
+            } else if (WIFSIGNALED(status)) {
+                char signame[SIG2STR_MAX];
+
+                if (sig2str(WTERMSIG(status), signame) < 0)
+                    sprintf(signame, "%d", WTERMSIG(status));
+                warn(_("%s terminated by signal %s.\n"), name, signame);
+            }
+
         }
         if (child < 0 && errno != ECHILD)
             warn(_("Cannot wait for processes - %s\n"), errstr);
@@ -828,6 +852,7 @@ main (int argc, char **argv)
     char *filename;
     char *tmpdir;
     struct sigaction sigact;
+    bool do_config = true;
 
     set_quoting_style(NULL, escape_quoting_style);
     
@@ -841,19 +866,23 @@ main (int argc, char **argv)
 #endif
 
     while (true) {
-	c = getopt_long(argc, argv, "", long_opts, NULL);
+	c = getopt_long(argc, argv, short_opts, long_opts, NULL);
 	if (c == -1)
 	    break;
 
 	switch (c) {
-	case HELP_OPT:
+	case 'n': /* --no-config */
+	    do_config = false;
+	    break;
+	case HELP_OPT: /* --help */
 	    printf(_("Usage: %s [OPTION]...\n"), quotearg(argv[0]));
 	    puts(_("Start microdc, a command-line based Direct Connect client.\n"));
-	    printf(_("      --help     display this help and exit\n"));
-	    printf(_("      --version  output version information and exit\n"));
+	    printf(_("  -n, --no-config  do not read config file on startup\n"));
+	    printf(_("      --help       display this help and exit\n"));
+	    printf(_("      --version    output version information and exit\n"));
 	    printf(_("\nReport bugs to <%s>.\n"), PACKAGE_BUGREPORT);
 	    exit(EXIT_SUCCESS);
-	case VERSION_OPT:
+	case VERSION_OPT: /* --version */
 	    version_etc(stdout, NULL, PACKAGE, VERSION, "Oskar Liljeblad", NULL);
 	    exit(EXIT_SUCCESS);
         default:
@@ -866,13 +895,19 @@ main (int argc, char **argv)
 	goto cleanup;
     }
 
+    main_process_id = getpid();
     sigact.sa_handler = signal_received;
     if (sigemptyset(&sigact.sa_mask) < 0) {
         warn(_("Cannot empty signal set - %s\n"), errstr);
 	goto cleanup;
     }
     sigact.sa_flags = SA_RESTART;
+#ifdef HAVE_STRUCT_SIGACTION_SA_RESTORER
     sigact.sa_restorer = NULL;
+#endif
+    /* Note: every signal registered with a non-ignore action here must
+     * also be registered in user.c, either with an action or as ignored.
+     */
     if (sigaction(SIGINT, &sigact, NULL) < 0) {
         warn(_("Cannot register signal handler - %s\n"), errstr);
 	goto cleanup;
@@ -938,9 +973,11 @@ main (int argc, char **argv)
     my_ul_slots = 3;
 
     command_init();
-    get_package_file("config", &filename);
-    run_script(filename, true);
-    free(filename);
+    if (do_config) {
+        get_package_file("config", &filename);
+        run_script(filename, true);
+        free(filename);
+    }      
 
     while (running) {
     	fd_set res_read_fds;
