@@ -1,6 +1,6 @@
 /* user.c - User communication (in separate process)
  *
- * Copyright (C) 2004-2005 Oskar Liljeblad
+ * Copyright (C) 2004, 2005 Oskar Liljeblad
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -12,9 +12,9 @@
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU Library General Public License for more details.
  *
- * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
+ * You should have received a copy of the GNU General Public License along
+ * with this program; if not, write to the Free Software Foundation,
+ * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
  */
 
 #include <config.h>
@@ -767,7 +767,7 @@ user_now_writable(DCUserConnLocal *ucl)
 {
     if (ucl->user_state == DC_USER_CONNECT) {
 	int error;
-        int size = sizeof(error);
+        socklen_t size = sizeof(error);
 
         if (getsockopt(ucl->user_socket, SOL_SOCKET, SO_ERROR, &error, &size) < 0) {
             warn(_("Cannot get error status - %s\n"), errstr);
@@ -937,11 +937,21 @@ user_main(int main_sockets[2], struct sockaddr_in *addr, int sock)
 	goto cleanup;
     }
     sigact.sa_flags = SA_RESTART;
+#ifdef HAVE_STRUCT_SIGACTION_SA_RESTORER
     sigact.sa_restorer = NULL;
+#endif
+    /* Note: every signal registered with a non-ignore action in main.c
+     * must also be registered here, either with an action or as ignored.
+     */
     if (sigaction(SIGTERM, &sigact, NULL) < 0 || sigaction(SIGUSR1, &sigact, NULL) < 0) {
         warn(_("Cannot register signal handler - %s\n"), errstr);
 	goto cleanup;
     }
+    sigact.sa_handler = SIG_IGN;
+    if (sigaction(SIGINT, &sigact, NULL) < 0 || sigaction(SIGCHLD, &sigact, NULL) < 0) {
+        warn(_("Cannot register signal handler - %s\n"), errstr);
+	goto cleanup;
+    }
 
     //ucl->user_recvq_last = 0;
     ucl->user_recvq = byteq_new(DEFAULT_RECVQ_SIZE);
