@@ -12,9 +12,9 @@
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU Library General Public License for more details.
  *
- * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
+ * You should have received a copy of the GNU General Public License along
+ * with this program; if not, write to the Free Software Foundation,
+ * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
  */
 
 #include <config.h>
@@ -710,7 +710,7 @@ cmd_set(int argc, char **argv)
 }
 
 void
-set_completion_selector(const char *line,int ws,int we,char *word,PtrV *results)
+set_command_completion_selector(const char *line,int ws,int we,char *word,PtrV *results)
 {
     char *name;
     DCVariable *var;
